package org.xbib.interlibrary.db.medea.postgresql;

import java.io.IOException;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.library.Role;
import org.xbib.interlibrary.api.order.BibliographicData;
import org.xbib.interlibrary.api.order.DeliveryScope;
import org.xbib.interlibrary.api.order.Order;
import org.xbib.interlibrary.api.order.Priority;
import org.xbib.interlibrary.api.order.Status;
import org.xbib.interlibrary.api.order.tracking.DisplayStyle;
import org.xbib.interlibrary.api.route.Route;
import org.xbib.interlibrary.db.AbstractOrderManager;
import org.xbib.interlibrary.db.DatabaseModule;
import org.xbib.interlibrary.db.table.OrderTable;
import org.xbib.interlibrary.db.table.SummaryOrderTable;
import org.xbib.interlibrary.db.table.TrackTable;
import org.xbib.jdbc.query.Database;
import org.xbib.jdbc.query.Table;

public class MedeaOrderManager extends AbstractOrderManager {

    private static final Logger logger = Logger.getLogger(MedeaOrderManager.class.getName());

    private final Map<String, Integer> orders;

    private final Map<String, Integer> stages;

    private final String readOrderInstId;

    private final String readOrderActionId;

    private final String writeOrderHist;

    private final String readMinStage;

    private final String readMaxStage;

    private final String readStage;

    private final String writeStageMovePos;

    private final String readDest;

    private final String writeOrderInsert;

    private final String readOrderNum;

    private final String writeStageInsert;

    private final String readInst;

    private final String writeOrderStatusUpdate;

    private final String writeOrderTimeUpdate;

    public MedeaOrderManager(DatabaseModule manager) throws IOException {
        super(manager);
        this.orders = new HashMap<>();
        this.stages = new HashMap<>();
        this.readOrderInstId = fromInputStream("read-order-inst-ids.sql");
        this.readOrderActionId = fromInputStream("read-order-action-id.sql");
        this.writeOrderHist = fromInputStream("write-order-hist.sql");
        this.readMinStage = fromInputStream("read-min-stage.sql");
        this.readMaxStage = fromInputStream("read-max-stage.sql");
        this.readStage = fromInputStream("read-stage.sql");
        this.writeStageMovePos = fromInputStream("write-stage-move-pos.sql");
        this.readDest = fromInputStream("read-dest.sql");
        this.writeOrderInsert = fromInputStream("write-order-insert.sql");
        this.readOrderNum = fromInputStream("read-order-num.sql");
        this.writeStageInsert = fromInputStream("write-stage-insert.sql");
        this.readInst = fromInputStream("read-inst.sql");
        this.writeOrderStatusUpdate = fromInputStream("write-order-status-update.sql");
        this.writeOrderTimeUpdate = fromInputStream("write-order-time-update.sql");
    }

    @Override
    public String createOrder(Order order) {
        if (!validate(order)) {
            return null;
        }
        Map<String, Object> orderMap = new HashMap<>();
        orderMap.put("transactionid", order.getRequester().getTransactionId());
        orderMap.put("clientcomment", order.getRequester().getClient().getClientComment());
        Map<String, Object> stageMap = new HashMap<>();
        stageMap.put("clientid", order.getRequester().getClient().getClientId());
        stageMap.put("clientemail", order.getRequester().getClient().getClientEmail());
        stageMap.put("clientgroup", order.getRequester().getClient().getClientGroup());
        stageMap.put("pickuplocation", order.getRequester().getClient().getPickupLocation());
        stageMap.put("pickupcode", order.getRequester().getClient().getPickupCode());
        stageMap.put("maxcost", order.getRequester().getClient().getMaxCost());
        stageMap.put("source", order.getBibliographicData().getSource());
        stageMap.put("sourceid", order.getBibliographicData().getSourceId());
        stageMap.put("year", order.getBibliographicData().getYear());
        stageMap.put("pagination", order.getBibliographicData().getPagination());
        stageMap.put("title", order.getBibliographicData().getTitle());
        stageMap.put("intitle", order.getBibliographicData().getInTitle());
        stageMap.put("author", order.getBibliographicData().getAuthor());
        stageMap.put("publisher", order.getBibliographicData().getPublisher());
        stageMap.put("publisherplace", order.getBibliographicData().getPublisherPlace());
        stageMap.put("uid", order.getRequester().getIsil());
        // volume and issue is a single field
        stageMap.put("volissue", createVolumeIssue(order));
        // collapse ISBN/ISSN
        if (order.getBibliographicData().getIsbn() != null) {
            stageMap.put("isn", order.getBibliographicData().getIsbn());
        }
        if (order.getBibliographicData().getIssn() != null) {
            stageMap.put("isn", order.getBibliographicData().getIssn());
        }
        // custom data which is printed on request form in delivery
        stageMap.put("clientcode", getClientCode(order));
        // set Medea-specific delivery code, derive it from requesterform delivery type
        stageMap.put("deliverycode", createDeliveryCode(getDeliveryStyle(order)));
        // Leihscheindruck
        Library library = databaseModule.getInterlibrary().getConfiguration().getLibrary(order.getRequester().getIsil());
        if (library != null) {
            stageMap.put("libraryname", library.getPostalAddress().getShortName());
            stageMap.put("libraryaddress", library.getPostalAddress().getStreet());
            stageMap.put("libraryaddress1", library.getPostalAddress().getZipCode() + " " + library.getPostalAddress().getPlace());
            stageMap.put("printedorder", createPrintedOrder(order));
        }
        // Anschriften
        if (!order.getRequester().getClient().getClientAddressLines().isEmpty()) {
            List<String> address = order.getRequester().getClient().getClientAddressLines();
            for (int i = 0; i < 3; i++) {
                stageMap.put("address" + i, i < address.size() ? address.get(i) : "");
            }
        } else {
            stageMap.put("address0", null);
            stageMap.put("address1", null);
            stageMap.put("address2", null);
        }
        if (orders.isEmpty()) {
            orders.putAll(databaseModule.getColumnSizesOfTable("orders"));
        }
        // map is now filled complety for size check
        truncate(orderMap, "transactionid", orders, "OR_TAN");
        truncate(orderMap, "clientcomment", orders, "OR_NOTE");
        if (stages.isEmpty()) {
            stages.putAll(databaseModule.getColumnSizesOfTable("stages"));
        }
        truncate(stageMap, "author", stages, "ST_ITM_ART_AUTHOR");
        truncate(stageMap, "title", stages, "ST_ITM_ART_TITLE");
        truncate(stageMap, "volissue", stages, "ST_ITM_VOL_ISSUE");
        truncate(stageMap, "pagination", stages, "ST_ITM_PAGINATION");
        truncate(stageMap, "intitle", stages, "ST_ITM_TITLE");
        truncate(stageMap, "publisher", stages, "ST_ITM_PUBLISHER");
        truncate(stageMap, "publisherplace", stages, "ST_ITM_PUBL_PLACE");
        truncate(stageMap, "isn", stages, "ST_ITM_ISN");
        truncate(stageMap, "source", stages, "ST_ITM_VER_REF_SRC");
        truncate(stageMap, "sourceid", stages, "ST_ITM_UNIQ_ID1");
        truncate(stageMap, "year", stages, "ST_ITM_PUBL_DATE_COMP");
        truncate(stageMap, "clientid", stages, "ST_BENNR");
        truncate(stageMap, "clientemail", stages, "ST_DEL_EMAIL");
        truncate(stageMap, "maxcost", stages, "ST_ZUSKOSTEN");
        truncate(stageMap, "clientcode", stages, "ST_LOCNAME");
        truncate(stageMap, "pickupcode", stages, "ST_DEL_FETCH_CODE");
        truncate(stageMap, "pickuplocation", stages, "ST_DEL_FETCH_LOCATION");
        truncate(stageMap, "uid", stages, "ST_HEIMSIG");
        truncate(stageMap, "libraryname", stages, "ST_HEIMNAME");
        truncate(stageMap, "libraryaddress", stages, "ST_HEIMADRESSE");
        truncate(stageMap, "libraryaddress1", stages, "ST_HEIMADRESSE2");
        truncate(stageMap, "printedorder", stages, "ST_RL_TEXT");
        truncate(stageMap, "address0", stages, "ST_DEL_NF_IM1");
        truncate(stageMap, "address1", stages, "ST_DEL_NF_IM2");
        truncate(stageMap, "address2", stages, "ST_DEL_NF_IM3"); // maximum of three address lines
        // check transactionID by select, we do not have and we do not want a DB constraint on tranasction ID
        String maybeOrdernumber = resolveTransactionId(order.getRequester().getTransactionId());
        if (maybeOrdernumber != null) {
            logger.log(Level.WARNING, "transaction ID " + order.getRequester().getTransactionId() + " already exists, rejecting");
            return null;
        }
        Table instTable = databaseModule.getSingleRow(readInst, Map.of("isil", order.getRequester().getIsil()));
        if (instTable.isEmpty()) {
            logger.log(Level.WARNING, "requester ISIL not in Medea insitutions " + order.getRequester().getIsil() + ", rejecting");
            return null;
        }
        BigDecimal instId = instTable.getValue(0, 0);
        orderMap.put("instid", instId);
        stageMap.put("instid", instId);
        // container class for lambda
        final OrderNumber orderNumber = new OrderNumber();
        databaseModule.transact(dbs -> {
            Database database = dbs.get();
            // first, insert order data, with fixed stage 1, state 10 ("undefined"), user 30 ("hbz")
            database.insert(writeOrderInsert, orderMap);
            // immediately read order num
            Table table = database.getSingleRow(readOrderNum, Map.of("transactionid", order.getRequester().getTransactionId()));
            if (!table.isEmpty()) {
                // read order number
                BigDecimal theOrderNumber = table.getValue(0, "ordernumber");
                orderNumber.setOrderNumber(theOrderNumber);
                stageMap.put("ordernumber", theOrderNumber);
                // write first stage (other stages are duplicated at routing time)
                database.insert(writeStageInsert, stageMap);
                // write history, break at 250 characters
                for (String wrappedline : createOriginOrder(order).split("(?<=\\G.{250})")) {
                    // 1 = "Bestellung wurde angelegt"
                    database.insert(writeOrderHist, Map.of("ordernumber", theOrderNumber, "instid", instId, "eventid", 1, "msg", wrappedline));
                }
            }
        });
        return orderNumber.getOrderNumber().toPlainString();
    }

    @Override
    public void updateOrder(Order order, Route route, String signComment) {
        String key = order.getOrderNumber();
        Objects.requireNonNull(key);
        Route oldRoute = databaseModule.getRouteManager().getRoute(key);
        Objects.requireNonNull(route);
        List<String> biblist = route.getEntries().stream().map(Route.Entry::getIsil).toList();
        final String biblistString = '-' + biblist.stream().limit(10).collect(Collectors.joining("-"));
        List<String> biblistalt = oldRoute != null ? oldRoute.getEntries().stream().map(Route.Entry::getIsil).toList() : List.of();
        final String biblistAltString = '*' + biblistalt.stream().limit(10).collect(Collectors.joining("*"));
        // check for bibliographic changes
        Map<String, Object> map = order.getBibliographicData().asMap().entrySet().stream()
                .filter(e -> allowedSignFields.containsKey(e.getKey()))
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
        // synthetic fields
        map.put("volissue", createVolumeIssue(order));
        BibliographicData data = getBibliographicData(key);
        Map<String, Object> dbmap = data.asMap().entrySet().stream()
                .filter(e -> allowedSignFields.containsKey(e.getKey()))
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
        final StringBuilder sb = new StringBuilder();
        dbmap.forEach((k, v) -> {
            if (map.containsKey(k)) {
                if (!map.get(k).equals(dbmap.get(k))) {
                    if (!sb.isEmpty()) {
                        sb.append(", ");
                    }
                    if (dbmap.containsKey(k)) {
                        sb.append("'").append(dbmap.get(k)).append("' geändert nach '").append(map.get(k)).append("'");
                    } else {
                        if (map.containsKey(k)) {
                            sb.append("neuer Wert '").append(map.get(k)).append("'");
                        }
                    }
                }
            } else {
                // if not present, add old value from db state
                if (v != null) {
                    map.put(k, v);
                }
            }
        });
        final String changes = sb.toString();
        databaseModule.transact(dbs -> {
            Database database = dbs.get();
            updateRoute(database, key, route, map, biblistString, biblistAltString);
            if (!changes.isEmpty()) {
                insertHistory(database, "setroute", "UPDATED", key, changes);
            }
            if (signComment != null) {
                insertHistory(database, "setroute", "NOTE", key, signComment);
            }
        });
    }

    @Override
    public void updateOrderStatus(Order order, Status status, String comment) {
        String statusCode = toStatusCode(status);
        if (statusCode == null) {
            logger.log(Level.WARNING, "status code can not be mapped from " + status);
            return;
        }
        databaseModule.transact(dbs -> {
            Database database = dbs.get();
            // update Medea order table
            database.update(writeOrderStatusUpdate, Map.of("key", order.getOrderNumber(), "status", statusCode));
            // important: update Medea stage table and timestamp
            database.update(writeOrderTimeUpdate, Map.of("key", order.getOrderNumber(), "status", statusCode));
            // history if status changes
            String oldStatus = order.getStatus().getStatusTranslation();
            String newStatus = toStatusTranslation(status);
            if (!oldStatus.equals(newStatus)) {
                String message = "Zustand geändert von '" + oldStatus + "' nach '" + newStatus + "'";
                if (comment != null && !comment.isEmpty()) {
                    message = message + " / " + comment;
                }
                insertHistory(database, "setstatus", "STATE_CHANGE", order.getOrderNumber(), message);
            }
        });
    }

    @Override
    public Order getOrder(String key) {
        String where = " where s.st_or_id = :key";
        String orderby = " order by e3.time desc, e2.time desc, e1.time desc";
        Table table = databaseModule.getSingleRow(readOrderSingleStatement + where + orderby, Map.of("key", key));
        return getOrder(table, 0);
    }

    public Table getOrderTable(String key) {
        return databaseModule.getSingleRow(readOrderSingleStatement + " where s.st_or_id = :key", Map.of("key", key));
    }

    public Table getOrderTable(String key, String isil) {
        return databaseModule.getSingleRow(readOrderSingleStatement + " where s.st_or_id = :key and (req.it_isil = :isil or sup.it_isil = :isil)", Map.of("key", key, "isil", isil));
    }


    protected Table trackOrder(String isil) {
        return databaseModule.getSingleRow(readOrderSingleStatement + " where req.it_isil = :isil", Map.of("isil", isil));
    }

    @Override
    protected Table getOrderHistoryTable(String key) {
        return databaseModule.getTable(readHistoryStatement + " where o.or_id = :key order by oh.oh_id asc",
                Map.of("key", key));
    }

    @Override
    protected void setTrackTableClauses(TrackTable table) {
        if (table.getWhere() == null) {
            table.setWhere(new ArrayList<>());
        }
        if (table.getParams() == null) {
            table.setParams(new HashMap<>());
        }
        if (table.getIsil() != null) {
            table.getWhere().add("req.it_isil = :isil");
            table.getParams().put("isil", table.getIsil());
        }
        if (table.getClientId() != null) {
            table.getWhere().add("s.st_bennr = :requesterid");
            table.getParams().put("requesterid", table.getClientId());
        }
        if (table.getCirculationSet() != null) {
            String s = table.getCirculationSet().stream().map(e -> switch (e) {
                case COPY -> "'K'";
                case FORM -> "'E'";
                default -> null;
            }).filter(Objects::nonNull).collect(Collectors.joining(","));
            if (!s.isEmpty()) {
                table.getWhere().add("s.st_type in (" + s + ")");
            }
        }
        if (table.getStatus() != null) {
            String s = String.join(",", translateOrderStatus(table.getStatus()));
            if (!s.isEmpty()) {
                table.getWhere().add("o.or_status in (" + s + ")");
            }
        }
        if (table.getFromLastActivityTime() != null && table.getToLastActivityTime() != null) {
            table.getParams().put("fromlastactivitytime", table.getFromLastActivityTime());
            table.getParams().put("tolastactivitytime", table.getToLastActivityTime());
            table.getWhere().add("s.st_when between :fromlastactivitytime and :tolastactivitytime");
        }
        table.setWhereClause(table.getWhere().isEmpty() ? "" : "where " + String.join(" and ", table.getWhere()));
        table.addSort("ordernumber", false);
    }

    @Override
    protected Table trackOrder(String isil, String transactionid) {
        return databaseModule.getSingleRow(readOrderSingleStatement +
                        " where o.or_tan = :transactionid" +
                        " and req.it_isil = :isil",
                Map.of("isil", isil, "transactionid", transactionid));
    }

    @Override
    protected Table trackOrder(String isil, String clientid, String transactionid) {
        return databaseModule.getSingleRow(readOrderSingleStatement +
                        " where s.st_bennr = :clientid" +
                        " and o.or_tan = :transactionid" +
                        " and req.it_isil = :isil",
                Map.of("isil", isil, "clientid", clientid, "transactionid", transactionid));
    }

    @Override
    protected Table trackOrder(String key, DisplayStyle displayStyle) {
        // limit history if for client
        String types = displayStyle == DisplayStyle.CLIENT ? "and h.oh_type in (1,2,4,6,9,11,33,42,43,44,97)" : "";
        return databaseModule.getTable(readHistoryStatement +
                        " where o.or_id = :key " + types + " order by oh.oh_id asc",
                Map.of("key", key));
    }

    @Override
    protected void setSummaryOrderTableClauses(SummaryOrderTable table) {
        if (table.getWhere() == null) {
            table.setWhere(new ArrayList<>());
        }
        if (table.getParams() == null) {
            table.setParams(new HashMap<>());
        }
        if (table.getRole() == Role.REQUESTER && table.getIsil() != null) {
            table.getWhere().add("req.it_isil = :isil");
            table.getParams().put("isil", table.getIsil());
        } else if (table.getRole() == Role.SUPPLIER && table.getIsil() != null) {
            table.getWhere().add("sup.it_isil = :isil");
            table.getParams().put("isil", table.getIsil());
        }
        if (table.getCirculationSet() != null) {
            String s = table.getCirculationSet().stream().map(e -> "'" + e.getMedeaCode() + "'")
                    .collect(Collectors.joining(","));
            if (!s.isEmpty()) {
                table.getWhere().add("s.st_type in (" + s + ")");
            }
        }
        if (table.getFromOrderTime() != null && table.getToOrderTime() != null) {
            table.getParams().put("fromordertime", table.getFromOrderTime());
            table.getParams().put("toordertime", table.getToOrderTime());
            table.getWhere().add("o.or_when between :fromordertime and :toordertime");
        }
        String where = table.getWhere().isEmpty() ? "" : "where " + String.join(" and ", table.getWhere());
        table.setStatement(String.format(table.getStatement(), where));
        table.addSort("circulation", true);
        table.addSort("status", true);
    }

    @Override
    protected void setOrderTableClauses(OrderTable table) {
        if (table.getWhere() == null) {
            table.setWhere(new ArrayList<>());
        }
        if (table.getParams() == null) {
            table.setParams(new HashMap<>());
        }
        if (table.getSearchParameters().getOrderSystem() != null) {
            table.getWhere().add("os.os_name = :ordersystem");
            table.getParams().put("ordersystem", table.getSearchParameters().getOrderSystem().name().toUpperCase(Locale.ROOT));
        }
        if (table.getSearchParameters().getCirculation() != null) {
            table.getWhere().add("s.st_type = :type");
            table.getParams().put("type", Character.toString(table.getSearchParameters().getCirculation().getMedeaCode()));
        }
        if (table.getSearchParameters().getPriority() != null) {
            for (Priority priority : table.getSearchParameters().getPriority()) {
                switch (priority) {
                    case LP1, LP3, LP4, LP5, LP6, LP7, LP8 -> {
                        // there are no such orders allowed in Medea
                    }
                    case UNDEFINED, LP2 -> table.getWhere().add("s.st_type in ('K', 'S')");
                    case LP9 -> table.getWhere().add("s.st_type = 'E'");
                }
            }
        }
        DeliveryScope deliveryScope = table.getSearchParameters().getDeliveryScope();
        if (deliveryScope != null) {
            if (deliveryScope == DeliveryScope.EXTERNAL) {
                table.getWhere().add("sup.it_type in ('V')");
            } else {
                table.getWhere().add("sup.it_type not in ('V')");
            }
        }
        if (table.getSearchParameters().getStatus() != null) {
            List<String> list = translateOrderStatus(table.getSearchParameters().getStatus());
            List<String> names = new ArrayList<>();
            int n = 1;
            for (String status : list) {
                names.add(":status" + n);
                table.getParams().put("status" + n, status);
                n++;
            }
            table.getWhere().add("o.or_status in (" + String.join(",", names) + ")");
        }
        Role role = table.getSearchParameters().getRole();
        if (role == Role.REQUESTER && table.getSearchParameters().getISIL() != null) {
            table.getWhere().add("req.it_isil = :isil");
            table.getParams().put("isil", table.getSearchParameters().getISIL());
        } else if (role == Role.SUPPLIER && table.getSearchParameters().getISIL() != null) {
            table.getWhere().add("sup.it_isil = :isil");
            table.getParams().put("isil", table.getSearchParameters().getISIL());
        } else if (role == Role.BOTH && table.getSearchParameters().getISIL() != null) {
            table.getWhere().add("sup.it_isil = :isil or req.it_isil = :isil");
            table.getParams().put("isil", table.getSearchParameters().getISIL());
        }
        if (table.getSearch() != null) {
            if (!table.getSearch().contains("%")) {
                table.getWhere().add("(o.or_id = :search or o.or_tan = :search)");
                table.getParams().put("search", table.getSearch());
            } else {
                table.getWhere().add("(o.or_id like :search or o.or_tan like :search)");
                table.getParams().put("search", table.getSearch());
            }
        }
        if (table.getFromOrderTime() != null && table.getToOrderTime() != null) {
            table.getParams().put("fromordertime", table.getFromOrderTime());
            table.getParams().put("toordertime", table.getToOrderTime());
            table.getWhere().add("o.or_when between :fromordertime and :toordertime");
        }
        if (table.getFromLastActivityTime() != null && table.getToLastActivityTime() != null) {
            table.getParams().put("fromlastactivitytime", table.getFromLastActivityTime());
            table.getParams().put("tolastactivitytime", table.getToLastActivityTime());
            table.getWhere().add("s.st_when between :fromlastactivitytime and :tolastactivitytime");
        }
        if (table.getFromSupplyTime() != null && table.getToSupplyTime() != null) {
            table.getParams().put("fromsupplytime", table.getFromSupplyTime());
            table.getParams().put("tosupplytime", table.getToSupplyTime());
            table.getWhere().add("s.st_when between :fromsupplytime and :tosupplytime");
        }
        if (table.getSearchParameters().getOffset() != null) {
            table.setOffset(table.getSearchParameters().getOffset());
        }
        if (table.getSearchParameters().getSize() != null) {
            table.setSize(table.getSearchParameters().getSize());
        }
        table.setWhereClause(table.getWhere().isEmpty() ? "" : "where " + String.join(" and ", table.getWhere()));
        logger.log(Level.FINE, "table params = " + table.getParams());
    }

    private boolean validate(Order order) {
        if (order.getRequester().getTransactionId() == null) {
            logger.log(Level.WARNING, "rejecting order creation without transaction ID: " + order.asMap());
            return false;
        }
        if (order.getBibliographicData().getTitle() == null) {
            logger.log(Level.WARNING, "rejecting order creation without title: " + order.asMap());
            return false;
        }
        if (order.getRequester().getClient().getClientId() == null) {
            logger.log(Level.WARNING, "rejecting order creation without client ID: " + order.asMap());
            return false;
        }
        if (order.getBibliographicData().getSource() == null) {
            logger.log(Level.WARNING, "rejecting order creation without source: " + order.asMap());
            return false;
        }
        if (order.getBibliographicData().getSourceId() == null) {
            logger.log(Level.WARNING, "rejecting order creation without source ID: " + order.asMap());
            return false;
        }
        if (order.getDeliveryStyle() == null) {
            logger.log(Level.WARNING, "rejecting order creation without delivery style: " + order.asMap());
            return false;
        }
        return true;
    }

    private String createVolumeIssue(Order order) {
        String volissue = null;
        BibliographicData requestForm = order.getBibliographicData();
        if (requestForm.getVolume() != null || requestForm.getIssue() != null) {
            StringBuilder sb = new StringBuilder();
            if (requestForm.getVolume() != null && !requestForm.getVolume().trim().isEmpty()) {
                sb.append("Band ").append(requestForm.getVolume());
            }
            if (requestForm.getIssue() != null && !requestForm.getIssue().trim().isEmpty()) {
                if (!sb.isEmpty()) {
                    sb.append(' ');
                }
                sb.append("Heft ").append(requestForm.getIssue());
            }
            volissue = sb.toString();
        }
        return volissue;
    }

    private String getClientCode(Order order) {
        StringBuilder sb = new StringBuilder();
        if (order.getRequester().getClient().getPickupCode() != null) {
            sb.append(order.getRequester().getClient().getPickupCode());
        }
        if (order.getRequester().getClient().getClientId() != null) {
            if (!sb.isEmpty()) {
                sb.append('/');
            }
            sb.append(order.getRequester().getClient().getClientId());
        }
        return sb.toString();
    }

    private String getDeliveryStyle(Order order) {
        return order.getDeliveryStyle().getLabel();
    }

    private String createDeliveryCode(String deliveryType) {
        String code;
        try {
            code = deliveryCodeBundle.getString(deliveryType);
        } catch (Exception e) {
            code = "6"; // PRINT
        }
        return code;
    }

    private void updateRoute(Database database,
                             String key,
                             Route route,
                             Map<String, Object> map,
                             String biblistString,
                             String biblistAltString) {
        List<String> routeKeys = route.getEntries().stream().map(Route.Entry::getIsil).toList();
        List<String> isils = new ArrayList<>();
        // get min stage for original delivery/notify type
        Table table = database.getSingleRow(readMinStage, Map.of("ordernumber", key));
        BigDecimal minDelType = table.getValue(0, "deltype");
        BigDecimal minNtfType = table.getValue(0, "ntftype");
        table = database.getSingleRow(readMaxStage, Map.of("ordernumber", key));
        BigDecimal n = table.getValue(0, "n");
        int m = 0;
        table = database.getSingleRow(readStage, Map.of("ordernumber", key, "n", n.intValue()));
        Map<String, Object> maxstage = new LinkedHashMap<>();
        table.getRow(maxstage, 0);
        // now move old seqno out of the way to position 0 to avoid duplicate row
        database.update(writeStageMovePos, Map.of("ordernumber", key, "m", m, "n", n.intValue()));
        int i = n.intValue();
        for (String isil : routeKeys) {
            Table destTable = database.getSingleRow(readDest, Map.of("isil", isil));
            BigDecimal dest = destTable.getValue(0, "id");
            if (dest == null) {
                logger.warning("unable to map " + isil + " to an institution, skipping");
            } else {
                isils.add(isil);
                // use maxstage as a template
                maxstage.put("ST_SEQNO", i); // number of stage
                maxstage.put("ST_TYPE", 'K'); // always 'K'
                maxstage.put("ST_STATUS", 10); // 10 = undefined
                maxstage.put("ST_DEST_INST", dest); // the target library
                maxstage.put("ST_ITM_SIGNATUR", buildSig(isil, route));
                truncate(maxstage, "ST_ITM_SIGNATUR", stages, "ST_ITM_SIGNATUR");
                maxstage.put("ST_WHEN", LocalDateTime.now()); // current timestamp
                // keep original notify/delivery type from min stage
                maxstage.put("ST_DEL_TYPE", minDelType.intValue());
                maxstage.put("ST_DEL_NF_TYPE", minNtfType.intValue());
                // write changed fields
                allowedSignFields.forEach((kk, vv) -> {
                    if (map.containsKey(kk)) {
                        maxstage.put(vv, map.get(kk));
                        truncate(maxstage, vv, stages, vv);
                    }
                });
                // delete sequence ID, must not be used
                maxstage.remove("ST_ID");
                String columnNames = String.join(",", maxstage.keySet());
                String bindNames = maxstage.keySet().stream().map(s -> ":" + s).collect(Collectors.joining(","));
                String statement = "insert into medea3.stages (ST_ID," + columnNames + ") values (stage_sq.nextval," + bindNames + ")";
                database.insert(statement, maxstage);
                i++;
            }
        }
        if (i > n.intValue()) {
            // adjust seqno to the found value, one after the updated route, and add biblist/biblistalt
            String statement = "update medea3.stages set st_seqno = :i, st_biblist = :biblist, st_biblistalt = :biblistalt where st_or_id = :key and st_seqno = :m";
            database.update(statement, Map.of("key", key, "m", m, "i", i, "biblist", biblistString, "biblistalt", biblistAltString));
            // reset route to begin of new route
            statement = "update medea3.orders set or_seqno = :n where or_id = :key";
            database.update(statement, Map.of("key", key, "n", n));
        } else {
            // nothing added, restore old seqno to n
            String statement = "update medea3.stages set st_seqno = :n where st_or_id = :key and st_seqno = :m";
            database.update(statement, Map.of("key", key, "m", m, "n", n));
        }
        String msg = isils.isEmpty() ? "kein Leitweg wurde gesetzt" : "Leitweg wurde ergänzt um " + String.join(", ", isils);
        insertHistory(database, "setroute", "LEITWEG_NEU", key, msg);
        setState(database, key, "Neu");
    }

    private void insertHistory(Database db, String type, String code, String ordernumber, String message) {
        Table table = db.getSingleRow(readOrderInstId, Map.of("ordernumber", ordernumber));
        BigDecimal supplierId = table.getValue(0, "supplier");
        BigDecimal requesterId = table.getValue(0, "requester");
        // depending on 'type', select requester or supplier id
        BigDecimal instid = switch (type) {
            case "requester", "view", "setroute" -> requesterId;
            default -> supplierId;
        };
        table = db.getSingleRow(readOrderActionId, Map.of("action", code));
        BigDecimal eventid = table.getValue(0, "eventid");
        // break into length of 250 to match maximum of 255
        if (instid != null && eventid != null && message != null) {
            String[] messages = message.split("(?<=\\G.{250})");
            for (String msg : messages) {
                if (msg != null && !msg.isEmpty()) {
                    Map<String, Object> map = Map.of("ordernumber", ordernumber,
                            "instid", instid.toPlainString(), "eventid", eventid.toPlainString(), "msg", msg);
                    logger.log(Level.INFO, "insert history: " + map);
                    db.insert(writeOrderHist, map);
                }
            }
        } else {
            logger.log(Level.WARNING, "unable to write, instid or eventid or message is null: code = " + code +
                    " eventid = " + eventid + " instid = " + instid + " message = " + message);
        }
    }

    private static String buildSig(String isil, Route route) {
        StringBuilder sb = new StringBuilder();
        sb.append("ISIL: ").append(isil);
        Optional<Collection<LibraryService>> collection = route.getEntries().stream()
                .filter(e -> e.getIsil().equals(isil))
                .map(Route.Entry::getServices).findFirst();
        if (collection.isEmpty()) {
            return sb.toString();
        }
        sb.append(" -");
        for (LibraryService libraryService : collection.get()) {
            String serviceISIL = libraryService.getISIL();
            if (!isil.equals(serviceISIL)) {
                sb.append(' ').append(serviceISIL).append(':');
            }
            List<LibraryService.Location> locations = libraryService.getLocations();
            for (LibraryService.Location location : locations) {
                String s = formatLocation(location);
                if (!s.isEmpty()) {
                    sb.append(' ').append(s);
                }
            }
        }
        return sb.toString();
    }

    private static String formatLocation(LibraryService.Location location) {
        if (location == null) {
            return "";
        }
        StringBuilder sb = new StringBuilder();
        String library = location.getLibrary();
        if (library != null) {
            sb.append(library);
        }
        String libraryName = location.getLibraryName();
        if (libraryName != null) {
            sb.append(" ").append(libraryName);
        }
        String collection = location.getCollection();
        if (collection != null) {
            if (!sb.isEmpty()) {
                sb.append(" ");
            }
            sb.append(collection);
        }
        String callnumber = location.getCallnumber();
        if (callnumber != null) {
            sb.append(" ").append(callnumber);
        }
        String description = location.getDescription();
        if (description != null) {
            sb.append(" ").append(description);
        }
        String textualHoldings = location.getTextualHoldings();
        if (textualHoldings != null) {
            sb.append(" ").append(textualHoldings);
        }
        List<LibraryService.LabeledUrl> list = location.getLinks();
        if (!list.isEmpty()) {
            for (LibraryService.LabeledUrl lu : list) {
                sb.append(" URL: ").append(lu.getUrl()).append(" (").append(lu.getLabel()).append(")");
            }
        }
        return sb.toString();
    }

    private static final Map<String, String> allowedSignFields = new HashMap<>() {
        {
            put("author", "ST_ITM_ART_AUTHOR");
            put("title", "ST_ITM_ART_TITLE");
            put("volissue", "ST_ITM_VOL_ISSUE");
            put("pagination", "ST_ITM_PAGINATION");
            put("intitle", "ST_ITM_TITLE");
            put("publisher", "ST_ITM_PUBLISHER");
            put("publisherplace", "ST_ITM_PUBL_PLACE");
            put("identifier", "ST_ITM_ISN");
            put("source", "ST_ITM_VER_REF_SRC");
            put("sourceid", "ST_ITM_UNIQ_ID1");
            put("year", "ST_ITM_PUBL_DATE_COMP");
        }
    };

    private void setState(Database database, String ordernumber, String newState) {
        String statement = "select ds_id from medea3.delstatus where ds_status = :state";
        Table table = databaseModule.getSingleRow(statement, Map.of("state", newState));
        BigDecimal state = table.getValue(0, "ds_id");
        if (state == null) {
            logger.warning("no state code found for new state " + newState);
            return;
        }
        statement = "select ds_status from medea3.delstatus, medea3.orders where ds_id = or_status and or_id = :ordernumber";
        table = databaseModule.getSingleRow(statement, Map.of("ordernumber", ordernumber));
        String oldState = table.getValue(0, "ds_status");
        if (oldState == null) {
            logger.warning("no state code found for old state " + oldState);
            return;
        }
        if (!oldState.equals(newState)) {
            // we MUST update twice, in orders and stages, because DB model is not normalized
            String stmt = "update medea3.orders set or_status = :state where or_id = :ordernumber";
            database.update(stmt, Map.of("ordernumber", ordernumber, "state", state));
            // we MUST update the activity time
            stmt = "update medea3.stages set st_status = :state, st_when = sysdate where st_id = (select s.st_id from  medea3.orders o  left join medea3.stages s on o.or_id = s.st_or_id and o.or_seqno = s.st_seqno  where  o.or_id = :ordernumber)";
            database.update(stmt, Map.of("ordernumber", ordernumber, "state", state));
            String msg = "alter Zustand '" + oldState + "', neuer Zustand '" + newState + "'";
            insertHistory(database, "requester", "UPDATED", ordernumber, msg);
        }
    }

    private static String createOriginOrder(Order order) {
        BibliographicData bibliographicData = order.getBibliographicData();
        StringBuilder sb = new StringBuilder();
        sb.append("Originalbestellung:");
        if (bibliographicData.getInTitle() != null) {
            sb.append(' ').append(bibliographicData.getInTitle());
        }
        if (bibliographicData.getPublisher() != null) {
            sb.append(", ").append(bibliographicData.getPublisher());
        }
        if (bibliographicData.getPublisherPlace() != null) {
            sb.append(", ").append(bibliographicData.getPublisherPlace());
        }
        if (bibliographicData.getIssn() != null) {
            sb.append(", ISSN ").append(bibliographicData.getIssn());
        }
        if (bibliographicData.getIsbn() != null) {
            sb.append(", ISBN ").append(bibliographicData.getIsbn());
        }
        if (bibliographicData.getTitle() != null) {
            sb.append(" Daraus: ");
            if (bibliographicData.getAuthor() != null) {
                sb.append(bibliographicData.getAuthor());
            }
            if (bibliographicData.getTitle() != null) {
                sb.append(", ").append(bibliographicData.getTitle());
            }
            if (bibliographicData.getVolume() != null) {
                sb.append(", Band ").append(bibliographicData.getVolume());
            }
            if (bibliographicData.getIssue() != null) {
                sb.append(", Heft ").append(bibliographicData.getIssue());
            }
            if (bibliographicData.getYear() != null) {
                sb.append(", ").append(bibliographicData.getYear());
            }
            if (bibliographicData.getPagination() != null) {
                sb.append(", ").append(bibliographicData.getPagination());
            }
        }
        if (order.getRequester().getClient().getClientComment() != null) {
            sb.append(" Bemerkung: ").append(order.getRequester().getClient().getClientComment());
        }
        if (order.getRequester().getClient().getPickupLocation() != null) {
            sb.append(" Abholort: ").append(order.getRequester().getClient().getPickupLocation());
        }
        return sb.toString();
    }

    private static String createPrintedOrder(Order order) {
        BibliographicData bibliographicData = order.getBibliographicData();
        StringBuilder sb = new StringBuilder();
        if (bibliographicData.getInTitle() != null) {
            sb.append(bibliographicData.getInTitle()).append("\n");
        }
        if (bibliographicData.getPublisher() != null) {
            sb.append(bibliographicData.getPublisher()).append("\n");
        }
        if (bibliographicData.getPublisherPlace() != null) {
            sb.append(bibliographicData.getPublisherPlace()).append("\n");
        }
        if (bibliographicData.getIssn() != null) {
            sb.append("ISSN ").append(bibliographicData.getIssn()).append("\n");
        }
        if (bibliographicData.getIsbn() != null) {
            sb.append("ISBN ").append(bibliographicData.getIsbn()).append("\n");
        }
        sb.append("Daraus:").append("\n");
        if (bibliographicData.getAuthor() != null) {
            sb.append(bibliographicData.getAuthor()).append("\n");
        }
        if (bibliographicData.getTitle() != null) {
            sb.append(bibliographicData.getTitle()).append("\n");
        }
        if (bibliographicData.getVolume() != null) {
            sb.append("Band ").append(bibliographicData.getVolume()).append(", ");
        }
        if (bibliographicData.getIssue() != null) {
            sb.append("Heft ").append(bibliographicData.getIssue()).append(", ");
        }
        if (bibliographicData.getYear() != null) {
            sb.append(bibliographicData.getYear()).append(", ");
        }
        if (bibliographicData.getPagination() != null) {
            sb.append("Seiten: ").append(bibliographicData.getPagination());
        }
        sb.append("\n");
        if (order.getRequester().getClient().getClientComment() != null) {
            sb.append("Bemerkung: ").append(order.getRequester().getClient().getClientComment()).append("\n");
        }
        if (order.getRequester().getClient().getPickupLocation() != null) {
            sb.append("Abholort: ").append(order.getRequester().getClient().getPickupLocation()).append("\n");
        }
        return sb.toString();
    }

    private static void truncate(Map<String, Object> map, String key, Map<String, Integer> columns, String columnName) {
        Object v = map.get(key);
        if (v == null) {
            return;
        }
        Integer size = columns.get(columnName);
        if (size == null) {
            logger.warning("undefined column: " + columnName);
            return;
        }
        if (v.toString().length() > size) {
            map.put(key, v.toString().substring(0, size));
            logger.warning("truncated: " + key + " larger than " + size + " -> " + map.get(key));
        }
    }

    static class OrderNumber {
        BigDecimal orderNumber;

        void setOrderNumber(BigDecimal ordernumber) {
            this.orderNumber = ordernumber;
        }

        BigDecimal getOrderNumber() {
            return orderNumber;
        }
    }

    private static List<String> translateOrderStatus(Status status) {
        List<String> list = new ArrayList<>();
        switch (status.getCode()) {
            case INCOMING -> {
                list.add("1"); // Neu
            }
            case PROCESSING -> {
                switch (status.getSubstatus()) {
                    case REQUESTFORM -> list.add("2"); // Deckblatt
                    case READY -> list.add("4"); // Gescannt
                    case TRANSFERRED -> {
                        list.add("14"); // Akzeptiert
                    }
                    case NOTTRANSFERRED -> {
                        list.add("16"); // Nicht akzeptiert
                    }
                    default -> {
                        list.add("3"); // Scan
                        list.add("4"); // Gescannt
                        list.add("14"); // Akzeptiert
                    }
                }
            }
            case SHIP -> {
                switch (status.getSubstatus()) {
                    case INPROCESS -> {
                        list.add("13"); // Druckvorbereitung
                    }
                    case READY -> {
                        list.add("18"); // Druckbereit
                    }
                    case PROBED -> {
                        list.add("5"); // Geliefert
                    }
                    default -> {
                        list.add("5"); // Geliefert
                        list.add("13"); // Druckvorbereitung
                        list.add("17"); // Versandbereit (obsolet!)
                        list.add("18"); // Druckbereit
                    }
                }
            }
            case IDLE -> {
                list.add("6"); // Rechnung
                list.add("10"); // Unbekannt
                list.add("11"); // Weiter
            }
            case NOTSUPPLIED -> {
                list.add("7"); // Abgesagt
                list.add("15"); // Unerledigt
            }
            case CANCELLED -> {
                list.add("8"); // Ungültig
            }
            case ERROR -> {
                list.add("9"); // Fehler
            }
            case SIGN -> {
                list.add("12"); // Signieren (nicht implementiert!)
            }
            // UNDEFINED, PREPARING, ROUTING, RETURN, END, EXPIRE, QUIT, INFO
        }
        return list;
    }

    private String toStatusCode(Status status) {
        String string = null;
        switch (status.getCode()) {
            case INCOMING -> {
                string = "1"; // Neu
            }
            case PROCESSING -> {
                switch (status.getSubstatus()) {
                    case REQUESTFORM -> string = "2"; // Deckblatt
                    case READY -> string = "4"; // Gescannt
                    case TRANSFERRED -> string = "14"; // Akzeptiert
                    case NOTTRANSFERRED -> string = "16"; // Nicht akzeptiert
                }
            }
            case SHIP -> {
                switch (status.getSubstatus()) {
                    case INPROCESS -> string = "13"; // Druckvorbereitung
                    case READY -> string = "18"; // Druckbereit
                    case PROBED -> string = "5"; // Geliefert
                    case NOTIFIED -> string = "5"; // Geliefert
                }
            }
            case NOTSUPPLIED -> string = "7"; // Abgesagt
            case CANCELLED -> string = "8"; // Ungültig
            case ERROR -> string = "9"; // Fehler
            case SIGN -> string = "12"; // Signieren (nicht implementiert!)
            default -> throw new UnsupportedOperationException();
            // unsupported: UNDEFINED, PREPARING, ROUTING, RETURN, END, EXPIRE, QUIT, INFO
        }
        return string;
    }

    public String toStatusTranslation(Status status) {
        String statusCode = toStatusCode(status);
        if (statusBundle.containsKey(statusCode)) {
            return statusBundle.getString(statusCode);
        } else {
            return status.getStatusText();
        }
    }
}
