package org.xbib.interlibrary.db.medea.postgresql;

import java.io.IOException;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.xbib.interlibrary.api.library.DefaultLibraryService;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.library.Mode;
import org.xbib.interlibrary.api.order.Circulation;
import org.xbib.interlibrary.api.route.Route;
import org.xbib.interlibrary.common.route.DefaultRoute;
import org.xbib.interlibrary.db.medea.postgresql.util.HttpAvail;
import org.xbib.interlibrary.http.HttpHelper;
import org.xbib.interlibrary.db.AbstractRouteManager;
import org.xbib.interlibrary.db.DatabaseModule;
import org.xbib.jdbc.query.Table;

public class MedeaRouteManager extends AbstractRouteManager {

    private static final Logger logger = Logger.getLogger(MedeaRouteManager.class.getName());

    private final HttpAvail httpAvail;

    public MedeaRouteManager(DatabaseModule databaseModule,
                             HttpHelper httpHelper) throws IOException {
        super(databaseModule);
        this.httpAvail = new HttpAvail(databaseModule.getInterlibrary().getSettings(), httpHelper);
    }

    @SuppressWarnings("unchecked")
    @Override
    protected DefaultRoute getRoute(Table table) {
        if (table.isEmpty()) {
            logger.log(Level.WARNING, "table is empty, no route");
            return null;
        }
        DefaultRoute route = new DefaultRoute();
        String requesterisil = table.getValue(0, "requesterisil");
        String source = table.getValue(0, "source");
        String sourceid = table.getValue(0, "sourceid");
        String yearString = table.getValue(0, "year");
        // clean year because it is user input (non numeric)
        yearString = cleanYear(yearString);
        // TODO use AvailRequest/AvailResponse via interlibrary HTTP client
        // TODO find out if print or online resource is correct
        boolean availSuccess = false;
        Map<String, Object> availResponse = null;
        Map<String, Object> availRequest = new LinkedHashMap<>();
        availRequest.put("requester", requesterisil);
        availRequest.put("source", source);
        availRequest.put("sourceid", sourceid);
        availRequest.put("year", yearString);
        availRequest.put("self_allowed", "true");
        availRequest.put("local_allowed", "true");
        availRequest.put("domain_allowed", "true");
        availRequest.put("enable_description", "true");
        availRequest.put("enable_link_priority", "true");
        availRequest.put("enable_print_resources", "true");
        availRequest.put("enable_online_resources", "false");
        try {
            availResponse = httpAvail.avail(availRequest);
            String availResponseStatus = (String) availResponse.get("status");
            availSuccess = "OK".equals(availResponseStatus);
        } catch (IOException e) {
            logger.log(Level.SEVERE, "unable to get avail response for request " + availRequest);
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        for (int i = 0; i < table.getRowCount(); i++) {
            BigDecimal pos = table.getValue(i, "pos");
            String active = table.getValue(i, "active");
            LocalDateTime time = table.getValue(i, "time");
            String transportservice = table.getValue(i, "transportservice");
            String circulationString = table.getValue(i, "circulation");
            Circulation circulation = Circulation.valueOf(circulationString);
            String supplierisil = table.getValue(i, "supplierisil");
            Library library = databaseModule.getInterlibrary().getConfiguration().getLibrary(supplierisil);
            if (library != null) {
                if (!supplierisil.startsWith("DE")) {
                    supplierisil = library.getISIL();
                }
                String name = library.getName();
                String domain = library.getDomain().getName();
                String organization = library.getDomain().getOrganization();
                boolean isOrganization = library.isOrganization();
                List<LibraryService> services = new ArrayList<>();
                if (availSuccess) {
                    String availResponseStatus = (String) availResponse.get("status");
                    if ("OK".equals(availResponseStatus)) {
                        Map<String, Object> meta = (Map<String, Object>) availResponse.get("meta");
                        if (meta != null) {
                            if (meta.containsKey("link")) {
                                route.setLinks((List<Map<String, Object>>) meta.get("link"));
                            }
                            if (meta.containsKey("linkedindicators")) {
                                route.setLinkedIndicators((List<Map<String, Object>>) meta.get("linkedindicators"));
                            }
                        }
                        if (availResponse.containsKey("interlibrary")) {
                            services.addAll(createServices(supplierisil, (Map<String, Object>) availResponse.get("interlibrary")));
                        }
                        if (availResponse.containsKey("interlibrarydomains")) {
                            services.addAll(createServices(supplierisil, (Map<String, Object>) availResponse.get("interlibrarydomains")));
                        }
                    } else {
                        logger.log(Level.WARNING, "interlibrary did not return OK: " + availResponseStatus);
                    }
                } else {
                    // read "Standort/Signatur" from Medea database and create a service
                    String supplierinfo = table.getValue(i, "supplierinfo");
                    if (supplierinfo != null) {
                        services.add(createService(pos.toString(), supplierisil, extractCollection(supplierinfo), extractCallnumber(supplierinfo)));
                    }
                }
                // ignore legacy form entry from Medea database, use only copy circulation
                if (circulation == Circulation.COPY) {
                    route.addEntry(pos.intValue(), "Y".equals(active), time, supplierisil, name, domain, organization,
                        isOrganization, transportservice, circulation, services);
                }
            }
        }
        createPrevious(route);
        createNext(route);
        return route;
    }

    private void createPrevious(DefaultRoute route) {
        List<String> list = new ArrayList<>();
        for (Route.Entry entry : route.getEntries()) {
            if (entry.isActive()) {
                break;
            }
            if (entry.isOrganization()) {
                list.add(entry.getOrganization());
            } else {
                list.add(entry.getIsil());
            }
        }
        route.setPrevious(list);
    }

    private void createNext(DefaultRoute route) {
        List<String> list = new ArrayList<>();
        boolean ok = false;
        for (Route.Entry entry : route.getEntries()) {
            if (ok) {
                if (entry.isOrganization()) {
                    list.add(entry.getOrganization());
                } else {
                    list.add(entry.getIsil());
                }
            }
            if (entry.isActive()) {
                ok = true;
            }
        }
        route.setNext(list);
    }

    /**
     * Clean year to integer digits, also clean characters for JSON.
     *
     * @param string the dirty input
     * @return cleaned year
     */
    private static String cleanYear(String string) {
        if (string == null) {
            return null;
        }
        String clean = string;
        clean = clean.replaceAll("[^\\d]", "").trim();
        if (clean.length() > 4) {
            clean = clean.substring(0, 4);
        }
        return clean;
    }

    private LibraryService createService(String pos, String isil, String collection, String callnumber) {
        DefaultLibraryService.DefaultLocation location = new DefaultLibraryService.DefaultLocation();
        location.setCollection(collection);
        location.setCallnumber(callnumber);
        return DefaultLibraryService.builder()
                .setId(pos)
                .setISIL(isil)
                .setServicePriority(2)
                .setMode(Mode.COPY)
                .setMaterials(List.of(Material.TEXT))
                .setLocations(List.of(location))
                .build();
    }

    private String extractCollection(String supplierInfo) {
        if (supplierInfo == null) {
            return null;
        }
        int pos = supplierInfo.indexOf("Standort:");
        String s = pos >= 0 ? supplierInfo.substring(pos + 9) : null;
        if (s == null) {
            return null;
        }
        // cut value by first blank
        pos = s.indexOf(' ');
        return pos >= 0 ? s.substring(0, pos) : s;
    }

    private String extractCallnumber(String supplierInfo) {
        if (supplierInfo == null) {
            return null;
        }
        int pos = supplierInfo.indexOf("Signatur:");
        String s = pos >= 0 ? supplierInfo.substring(pos + 9) : supplierInfo;
        Matcher matcher = callnumberPattern.matcher(s);
        if (matcher.find()) {
            return matcher.group(1);
        }
        return s;
    }

    private static final Pattern callnumberPattern = Pattern.compile(">(.*?)<");
}
